import typing
import numpy as np
import open3d as o3d


PointCloud = o3d.geometry.PointCloud
KDFlann = o3d.geometry.KDTreeFlann


def get_neighbour_cloud(
    iter_cloud: o3d.geometry.PointCloud,
    search_cloud: o3d.geometry.PointCloud,
    kdtree: o3d.geometry.KDTreeFlann,
    n: int,
) -> typing.Tuple[o3d.geometry.PointCloud, np.ndarray]:
    def get_neighbour(
        point: np.ndarray,
        kdtree: o3d.geometry.KDTreeFlann,
        n: int,
    ) -> np.ndarray:
        rpoint = point.reshape((3, 1))
        [_, idx, dists] = kdtree.search_knn_vector_3d(rpoint, n + 1)
        return np.array((idx[-1], dists[-1]))

    def finder(point: np.ndarray):
        return get_neighbour(point, kdtree, n)

    [idxs, sqrdists] = np.apply_along_axis(
        finder,
        axis=1,
        arr=iter_cloud.points,
    ).T
    idxs = idxs.astype(int)
    neigh_points = np.take(search_cloud.points, idxs, axis=0)
    neigh_cloud = o3d.geometry.PointCloud()
    neigh_cloud.points = o3d.utility.Vector3dVector(neigh_points)

    if search_cloud.has_colors():
        neigh_colors = np.take(search_cloud.colors, idxs, axis=0)
        neigh_cloud.colors = o3d.utility.Vector3dVector(neigh_colors)

    return (neigh_cloud, sqrdists)


class CloudPair:
    clouds: typing.Tuple[PointCloud, PointCloud]
    _trees: typing.Tuple[typing.Optional[KDFlann], typing.Optional[KDFlann]]
    _neigh_clouds: typing.Tuple[PointCloud, PointCloud]
    _neigh_dists: typing.Tuple[
        typing.Optional[np.ndarray],
        typing.Optional[np.ndarray],
    ]

    def __init__(
        self,
        origin_cloud: o3d.geometry.PointCloud,
        reconst_cloud: o3d.geometry.PointCloud,
    ):
        self.clouds = (origin_cloud, reconst_cloud)

        if not self.clouds[0].has_normals():
            self.clouds[0].estimate_normals()
        if not self.clouds[1].has_normals():
            self.clouds[1].estimate_normals()
        self._trees = tuple(map(KDFlann, self.clouds))

        origin_neigh_cloud, origin_neigh_dists = get_neighbour_cloud(
            iter_cloud=self.clouds[0],
            search_cloud=self.clouds[1],
            kdtree=self._trees[1],
            n=0,
        )
        reconst_neigh_cloud, reconst_neigh_dists = get_neighbour_cloud(
            iter_cloud=self.clouds[1],
            search_cloud=self.clouds[0],
            kdtree=self._trees[0],
            n=0,
        )
        self._neigh_clouds = (origin_neigh_cloud, reconst_neigh_cloud)
        self._neigh_dists = (origin_neigh_dists, reconst_neigh_dists)

    @property
    def origin_cloud(self):
        return self.clouds[0]

    @property
    def reconst_cloud(self):
        return self.clouds[1]

    def get_left_error_vector(self):
        return np.subtract(
            self.clouds[0].points,
            self._neigh_clouds[0].points,
        )

    def get_right_error_vector(self):
        return np.subtract(
            self.clouds[1].points,
            self._neigh_clouds[1].points,
        )

    def get_left_neighbour_distances(self):
        return self._neigh_dists[0]

    def get_right_neighbour_distances(self):
        return self._neigh_dists[1]

    def get_boundary_sqrt_distances(self):
        return self.clouds[0].compute_nearest_neighbor_distance()

    def get_extent(self):
        return self.clouds[0].get_minimal_oriented_bounding_box().extent

    def get_left_colors(self):
        return self.clouds[0].colors

    def get_right_colors(self):
        return self.clouds[1].colors

    def get_left_neighbour_colors(self):
        return self._neigh_clouds[0].colors

    def get_right_neighbour_colors(self):
        return self._neigh_clouds[1].colors



import typing
import pandas as pd
from .metric import (
    AbstractMetric,
    SymmetricMetric,
    PrimaryMetric,
    SecondaryMetric,
)
from .logger import get_logger
from .cloud_pair import CloudPair

logger = get_logger()


class CalculateResult:
    _metrics: typing.List[AbstractMetric]

    def __init__(self, metrics: typing.List[AbstractMetric]):
        self._metrics = metrics

    def as_dict(self) -> typing.Dict[str, typing.Any]:
        d = dict()
        for metric in self._metrics:
            d[metric._key()] = metric.value
        return d

    def as_df(self) -> pd.DataFrame:
        # metrics = [str(metric) for metric in self._metrics]
        metric_dict = {
            "label": [],
            "is_left": [],
            "point-to-plane": [],
            "value": [],
        }

        for metric in self._metrics:
            label = metric.__class__.__name__
            if isinstance(metric, SymmetricMetric):
                child_label = metric.metrics[0].__class__.__name__
                label = child_label + "(symmetric)"
            metric_dict["label"].append(label)
            is_left = ""
            if hasattr(metric, "is_left"):
                is_left = metric.is_left
            metric_dict["is_left"].append(is_left)
            point_to_plane = ""
            if hasattr(metric, "point_to_plane"):
                point_to_plane = metric.point_to_plane
            metric_dict["point-to-plane"].append(point_to_plane)
            metric_dict["value"].append(str(metric.value))

        return pd.DataFrame(metric_dict)

    def __str__(self) -> str:
        return str(self.as_df())


class MetricCalculator:
    _cloud_pair: CloudPair
    _calculated_metrics: typing.Dict[typing.Tuple, 'AbstractMetric'] = {}

    def __init__(self, cloud_pair: CloudPair):
        self._cloud_pair = cloud_pair

    def _metric_recursive_calculate(
        self,
        metric: 'AbstractMetric',
    ) -> 'AbstractMetric':
        if metric._key() in self._calculated_metrics:
            return self._calculated_metrics[metric._key()]

        if not isinstance(metric, (PrimaryMetric, SecondaryMetric)):
            # Would this work?
            unknwn = metric.__class__.__name__
            raise RuntimeError(
                f"Metric of unknown AbstractMetric subclass {unknwn}",
            )

        if isinstance(metric, PrimaryMetric):
            metric = typing.cast(PrimaryMetric, metric)
            metric.calculate(self._cloud_pair)
            self._calculated_metrics[metric._key()] = metric
            return metric

        calculated_deps = {}
        for dep_key, dep_metric in metric._get_dependencies().items():
            calculated_dep_metric = self._metric_recursive_calculate(
                metric=dep_metric,
            )
            calculated_deps[dep_key] = calculated_dep_metric

        metric.calculate(**calculated_deps)
        self._calculated_metrics[metric._key()] = metric

        return metric

    def calculate(
        self,
        metrics_list: typing.List[AbstractMetric],
    ) -> CalculateResult:
        # logger.info("%s metrics to calculate", len(metrics_list))

        calculated_metrics_list = []
        for metric in metrics_list:
            calculated_metric = self._metric_recursive_calculate(metric)
            calculated_metrics_list.append(calculated_metric)

        return CalculateResult(calculated_metrics_list)



import logging
import sys


def get_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



import click


@click.command()
@click.option(
    "--ocloud",
    required=True,
    type=str,
    help="Original point cloud.",
)
@click.option(
    "--pcloud",
    required=True,
    type=str,
    help="Processed point cloud.",
)
@click.option(
    "--color",
    required=False,
    type=click.Choice(["rgb", "ycc"]),
    help="Report color distortions as well.",
)
@click.option(
    "--hausdorff",
    required=False,
    is_flag=True,
    help=" ".join((
        "Report hausdorff metric as well. If --point-to-plane is provided,",
        "then hausdorff point-to-plane would be reported too",
        )),
)
@click.option(
    "--point-to-plane",
    required=False,
    is_flag=True,
    help="Report point-to-plane distance as well.",
)
@click.option(
    "--csv",
    required=False,
    is_flag=True,
    help="Print output in csv format.",
)
def cli(
    ocloud: str,
    pcloud: str,
    color: str,
    hausdorff: bool,
    point_to_plane: bool,
    csv: bool,
) -> None:
    import open3d as o3d
    from .cloud_pair import CloudPair
    from .calculator import MetricCalculator
    from .options import CalculateOptions, transform_options

    ocloud_cloud, pcloud_cloud = map(o3d.io.read_point_cloud, (ocloud, pcloud))
    cloud_pair = CloudPair(ocloud_cloud, pcloud_cloud)
    calculator = MetricCalculator(cloud_pair)
    options = CalculateOptions(
        color=color,
        hausdorff=hausdorff,
        point_to_plane=point_to_plane,
    )
    metrics = transform_options(options)
    result = calculator.calculate(metrics).as_df()

    if csv:
        print(result.to_csv())
    else:
        print(result.to_string())



import typing
from .metric import (
    AbstractMetric,
    MinSqrtDistance,
    MaxSqrtDistance,
    GeoMSE,
    GeoPSNR,
    ColorMSE,
    ColorPSNR,
    GeoHausdorffDistance,
    GeoHausdorffDistancePSNR,
    SymmetricMetric,
)


class CalculateOptions:
    color: typing.Optional[str]
    hausdorff: bool
    point_to_plane: bool

    def __init__(
        self,
        color: typing.Optional[str] = None,
        hausdorff: bool = False,
        point_to_plane: bool = False,
    ):
        self.color = color
        self.hausdorff = hausdorff
        self.point_to_plane = point_to_plane


def transform_options(
    options: 'CalculateOptions',
) -> typing.List['AbstractMetric']:
    metrics = [
        MinSqrtDistance(),
        MaxSqrtDistance(),
        GeoMSE(is_left=True, point_to_plane=False),
        GeoMSE(is_left=False, point_to_plane=False),
        SymmetricMetric(
            metrics=(
                GeoMSE(is_left=True, point_to_plane=False),
                GeoMSE(is_left=False, point_to_plane=False),
            ),
            is_proportional=False,
        ),
        GeoPSNR(is_left=True, point_to_plane=False),
        GeoPSNR(is_left=False, point_to_plane=False),
        SymmetricMetric(
            metrics=(
                GeoPSNR(is_left=True, point_to_plane=False),
                GeoPSNR(is_left=False, point_to_plane=False),
            ),
            is_proportional=True,
        ),
    ]

    if (
        # self.clouds[0].has_colors() and
        # self.clouds[1].has_colors() and
        (options.color is not None)
    ):
        metrics += [
            ColorMSE(is_left=True, color_scheme=options.color),
            ColorMSE(is_left=False, color_scheme=options.color),
            SymmetricMetric(
                metrics=(
                    ColorMSE(is_left=True, color_scheme=options.color),
                    ColorMSE(is_left=False, color_scheme=options.color),
                ),
                is_proportional=False,
            ),
            ColorPSNR(is_left=True, color_scheme=options.color),
            ColorPSNR(is_left=False, color_scheme=options.color),
            SymmetricMetric(
                metrics=(
                    ColorPSNR(is_left=True, color_scheme=options.color),
                    ColorPSNR(is_left=False, color_scheme=options.color),
                ),
                is_proportional=True,
            ),
        ]

    if options.point_to_plane:
        metrics += [
            GeoMSE(is_left=True, point_to_plane=True),
            GeoMSE(is_left=False, point_to_plane=True),
            SymmetricMetric(
                metrics=(
                    GeoMSE(is_left=True, point_to_plane=True),
                    GeoMSE(is_left=False, point_to_plane=True),
                ),
                is_proportional=False,
            ),
            GeoPSNR(is_left=True, point_to_plane=True),
            GeoPSNR(is_left=False, point_to_plane=True),
            SymmetricMetric(
                metrics=(
                    GeoPSNR(is_left=True, point_to_plane=True),
                    GeoPSNR(is_left=False, point_to_plane=True),
                ),
                is_proportional=True,
            ),
        ]

    if options.hausdorff:
        metrics += [
            GeoHausdorffDistance(is_left=True, point_to_plane=False),
            GeoHausdorffDistance(is_left=False, point_to_plane=False),
            SymmetricMetric(
                metrics=(
                    GeoHausdorffDistance(
                        is_left=True,
                        point_to_plane=False,
                    ),
                    GeoHausdorffDistance(
                        is_left=False,
                        point_to_plane=False,
                    ),
                ),
                is_proportional=False,
            ),
            GeoHausdorffDistancePSNR(is_left=True, point_to_plane=False),
            GeoHausdorffDistancePSNR(is_left=False, point_to_plane=False),
            SymmetricMetric(
                metrics=(
                    GeoHausdorffDistancePSNR(
                        is_left=True,
                        point_to_plane=False,
                    ),
                    GeoHausdorffDistancePSNR(
                        is_left=False,
                        point_to_plane=False,
                    ),
                ),
                is_proportional=True,
            ),
        ]

    if options.hausdorff and options.point_to_plane:
        metrics += [
            GeoHausdorffDistance(is_left=True, point_to_plane=True),
            GeoHausdorffDistance(is_left=False, point_to_plane=True),
            GeoHausdorffDistancePSNR(is_left=True, point_to_plane=True),
            GeoHausdorffDistancePSNR(is_left=False, point_to_plane=True),
            SymmetricMetric(
                metrics=(
                    GeoHausdorffDistance(
                        is_left=True,
                        point_to_plane=True,
                    ),
                    GeoHausdorffDistance(
                        is_left=False,
                        point_to_plane=True,
                    ),
                ),
                is_proportional=False,
            ),
            SymmetricMetric(
                metrics=(
                    GeoHausdorffDistancePSNR(
                        is_left=True,
                        point_to_plane=True,
                    ),
                    GeoHausdorffDistancePSNR(
                        is_left=False,
                        point_to_plane=True,
                    ),
                ),
                is_proportional=True,
            ),
        ]

    return metrics



import typing
import abc
import numpy as np
import open3d as o3d
from .logger import get_logger
from .cloud_pair import CloudPair

logger = get_logger()

PointCloud = o3d.geometry.PointCloud
KDFlann = o3d.geometry.KDTreeFlann


class AbstractMetric(abc.ABC):
    value: typing.Any

    def _key(self) -> typing.Tuple:
        return (self.__class__.__name__,)

    @abc.abstractmethod
    def calculate(
        self,
        cloud_pair: CloudPair,
        **kwargs: typing.Dict[str, 'AbstractMetric']
    ) -> None:
        raise NotImplementedError("calculate is not implemented")

    def __str__(self) -> str:
        return "{key}: {value}".format(key=self._key(), value=str(self.value))


class PrimaryMetric(AbstractMetric):
    @abc.abstractmethod
    def calculate(
        self,
        cloud_pair: CloudPair,
    ) -> None:
        raise NotImplementedError("calculate is not implemented")


class SecondaryMetric(AbstractMetric):
    def _get_dependencies(self) -> typing.Dict[str, 'AbstractMetric']:
        return {}

    @abc.abstractmethod
    def calculate(
        self,
        **kwargs: typing.Dict[str, 'AbstractMetric']
    ) -> None:
        raise NotImplementedError("calculate is not implemented")


class DirectionalMetric(AbstractMetric):
    is_left: bool

    def __init__(self, is_left: bool):
        self.is_left = is_left

    def _key(self) -> typing.Tuple:
        return super()._key() + (self.is_left,)


class PointToPlaneable(DirectionalMetric):
    point_to_plane: bool

    def __init__(self, is_left: bool, point_to_plane: bool):
        super().__init__(is_left)
        self.point_to_plane = point_to_plane

    def _key(self) -> typing.Tuple:
        return super()._key() + (self.point_to_plane,)


class PrimaryErrorVector(PrimaryMetric, DirectionalMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        if self.is_left:
            self.value = cloud_pair.get_left_error_vector()
            return

        self.value = cloud_pair.get_right_error_vector()


class NeighbourDistances(PrimaryMetric, DirectionalMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        if self.is_left:
            self.value = cloud_pair.get_left_neighbour_distances()
            return

        self.value = cloud_pair.get_right_neighbour_distances()


class CloudNormals(PrimaryMetric, DirectionalMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        if self.is_left:
            self.value = np.asarray(cloud_pair.clouds[0].normals)
            return

        self.value = np.asarray(cloud_pair.clouds[1].normals)


class CloudExtent(PrimaryMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        self.value = cloud_pair.get_extent()


class CloudColors(PrimaryMetric, DirectionalMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        if self.is_left:
            self.value = cloud_pair.get_left_colors()
            return

        self.value = cloud_pair.get_right_colors()


class NeighbourColors(PrimaryMetric, DirectionalMetric):
    def calculate(self, cloud_pair: CloudPair) -> None:
        if self.is_left:
            self.value = cloud_pair.get_left_neighbour_colors()
            return

        self.value = cloud_pair.get_right_neighbour_colors()


class ErrorVector(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        deps = {
            "primary_error_vector": PrimaryErrorVector(is_left=self.is_left)
        }
        if self.point_to_plane:
            deps["cloud_normals"] = CloudNormals(is_left=not self.is_left)
        return deps

    def calculate(
        self,
        primary_error_vector: PrimaryErrorVector,
        cloud_normals: typing.Optional[CloudNormals] = None,
    ) -> None:
        if not self.point_to_plane:
            self.value = np.apply_along_axis(
                func1d=np.linalg.norm,
                axis=1,
                arr=primary_error_vector.value,
            )
            return

        error_shape = primary_error_vector.value.shape
        plane_errs = np.zeros(shape=(error_shape[0],))
        for i in range(error_shape[0]):
            plane_errs[i] = np.dot(
                primary_error_vector.value[i],
                cloud_normals.value[i],
            )
        self.value = plane_errs


class EuclideanDistance(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        if not self.point_to_plane:
            return {
                "neighbour_distances": NeighbourDistances(is_left=self.is_left)
            }

        return {
            "error_vector": ErrorVector(
                is_left=self.is_left,
                point_to_plane=self.point_to_plane,
            )
        }

    def calculate(
        self,
        neighbour_distances: typing.Optional[NeighbourDistances] = None,
        error_vector: typing.Optional[ErrorVector] = None,
    ) -> None:
        if not self.point_to_plane:
            self.value = neighbour_distances.value
            return

        self.value = np.square(error_vector.value)


class BoundarySqrtDistances(PrimaryMetric):
    def calculate(
        self,
        cloud_pair: CloudPair,
    ) -> None:
        inner_dists = cloud_pair.get_boundary_sqrt_distances()
        self.value = (np.min(inner_dists), np.max(inner_dists))


class MinSqrtDistance(SecondaryMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {"boundary_metric": BoundarySqrtDistances()}

    def calculate(
        self,
        boundary_metric: BoundarySqrtDistances,
    ) -> None:
        self.value = boundary_metric.value[0]


class MaxSqrtDistance(SecondaryMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {"boundary_metric": BoundarySqrtDistances()}

    def calculate(
        self,
        boundary_metric: BoundarySqrtDistances,
    ) -> None:
        self.value = boundary_metric.value[1]


class GeoMSE(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "euclidean_distance": EuclideanDistance(
                is_left=self.is_left,
                point_to_plane=self.point_to_plane,
            )
        }

    def calculate(
        self,
        euclidean_distance: EuclideanDistance,
    ) -> None:
        n = euclidean_distance.value.shape[0]
        sse = np.sum(euclidean_distance.value, axis=0)
        self.value = sse / n


class GeoPSNR(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "cloud_extent": CloudExtent(),
            "geo_mse": GeoMSE(
                is_left=self.is_left,
                point_to_plane=self.point_to_plane,
            ),
        }

    def calculate(
        self,
        cloud_extent: CloudExtent,
        geo_mse: GeoMSE,
    ) -> None:
        peak = np.max(cloud_extent.value)
        self.value = 10 * np.log10(peak**2 / geo_mse.value)


class ColorMetric(DirectionalMetric):
    color_scheme: str

    def __init__(self, is_left: bool, color_scheme: str):
        super().__init__(is_left)
        self.color_scheme = color_scheme

    def _key(self) -> typing.Tuple:
        return super()._key() + (self.color_scheme,)


def transform_colors(
    colors: np.ndarray,
    source_scheme: str,
    target_scheme: str,
) -> np.ndarray:
    if source_scheme == target_scheme:
        return colors

    transform = None
    if (source_scheme == "rgb") and (target_scheme == "ycc"):
        transform = np.array([
            [0.2126, 0.7152, 0.0722],
            [-0.1146, -0.3854, 0.5],
            [0.5, -0.4542, -0.0458],
        ])
    if (source_scheme == "rgb") and (target_scheme == "yuv"):
        transform = np.array([
            [0.25, 0.5, 0.25],
            [1, 0, -1],
            [-0.5, 1, -0.5]
        ])

    def converter(c: np.ndarray) -> np.ndarray:
        return np.matmul(transform , c)

    return np.apply_along_axis(
        func1d=converter,
        axis=1,
        arr=colors,
    )


def get_color_peak(color_scheme: str) -> np.float64:
    colors_to_values = {
        "rgb": 255.0,
        "ycc": 1.0,
        "yuv": 1.0,
    }
    return colors_to_values[color_scheme]


class ColorMSE(SecondaryMetric, ColorMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "origin_cloud_colors": CloudColors(is_left=self.is_left),
            "neighbour_cloud_colors": NeighbourColors(is_left=self.is_left),
        }

    def calculate(
        self,
        origin_cloud_colors: CloudColors,
        neighbour_cloud_colors: NeighbourColors,
    ) -> None:
        orig_colors = np.copy(origin_cloud_colors.value)
        neigh_colors = np.copy(neighbour_cloud_colors.value)

        orig_colors = transform_colors(
            colors=orig_colors,
            source_scheme="rgb",
            target_scheme=self.color_scheme,
        )

        neigh_colors = transform_colors(
            colors=neigh_colors,
            source_scheme="rgb",
            target_scheme=self.color_scheme,
        )

        diff = np.subtract(
            orig_colors,
            neigh_colors,
        )
        self.value = np.mean(diff**2, axis=0)


class ColorPSNR(SecondaryMetric, ColorMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "color_mse": ColorMSE(
                is_left=self.is_left,
                color_scheme=self.color_scheme,
            ),
        }

    def calculate(
        self,
        color_mse: ColorMSE,
    ) -> None:
        peak = get_color_peak(self.color_scheme)
        self.value = 10 * np.log10(peak**2 / color_mse.value)


class GeoHausdorffDistance(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "euclidean_distance": EuclideanDistance(
                is_left=self.is_left,
                point_to_plane=self.point_to_plane,
            )
        }

    def calculate(
        self,
        euclidean_distance: EuclideanDistance,
    ) -> None:
        self.value = np.max(euclidean_distance.value, axis=0)


class GeoHausdorffDistancePSNR(SecondaryMetric, PointToPlaneable):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "max_sqrt": MaxSqrtDistance(),
            "hausdorff_distance": GeoHausdorffDistance(
                is_left=self.is_left,
                point_to_plane=self.point_to_plane,
            ),
        }

    def calculate(
        self,
        max_sqrt: MaxSqrtDistance,
        hausdorff_distance: GeoHausdorffDistance,
    ) -> None:
        self.value = 10 * np.log10(
            max_sqrt.value**2 / hausdorff_distance.value,
        )


class ColorHausdorffDistance(SecondaryMetric, ColorMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "origin_cloud_colors": CloudColors(is_left=self.is_left),
            "neighbour_cloud_colors": NeighbourColors(is_left=self.is_left),
        }

    def calculate(
        self,
        origin_cloud_colors: CloudColors,
        neighbour_cloud_colors: NeighbourColors,
    ) -> None:
        orig_colors = np.copy(origin_cloud_colors.value)
        neigh_colors = np.copy(neighbour_cloud_colors.value)

        orig_colors = transform_colors(
            colors=orig_colors,
            source_scheme="rgb",
            target_scheme=self.color_scheme,
        )

        neigh_colors = transform_colors(
            colors=neigh_colors,
            source_scheme="rgb",
            target_scheme=self.color_scheme,
        )

        diff = np.subtract(
            orig_colors,
            neigh_colors,
        )

        # ???
        if self.color_scheme == "rgb":
            rgb_scale = 255
            diff = rgb_scale * diff

        self.value = np.max(diff**2, axis=0)


class ColorHausdorffDistancePSNR(SecondaryMetric, ColorMetric):
    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "hausdorff_distance": ColorHausdorffDistance(
                is_left=self.is_left,
                color_scheme=self.color_scheme,
            ),
        }

    def calculate(
        self,
        hausdorff_distance: ColorHausdorffDistance,
    ) -> None:
        peak = get_color_peak(self.color_scheme)
        self.value = 10 * np.log10(peak**2 / hausdorff_distance.value)


class SymmetricMetric(SecondaryMetric):
    is_proportional: bool
    metrics: typing.List[DirectionalMetric]

    def _get_dependencies(self) -> typing.Dict[str, AbstractMetric]:
        return {
            "lmetric": self.metrics[0],
            "rmetric": self.metrics[1],
        }

    def __init__(
        self,
        metrics: typing.List[DirectionalMetric],
        is_proportional: bool,
    ):
        if len(metrics) != 2:
            raise ValueError("Must be exactly two metrics")
        if metrics[0].__class__ != metrics[1].__class__:
            lclass = metrics[0].__class__
            rclass = metrics[1].__class__
            raise ValueError(
                f"Metrics must be of same class, got: {lclass}, {rclass}"
            )
        self.metrics = metrics
        self.is_proportional = is_proportional

    def _key(self) -> typing.Tuple:
        return super()._key() + self.metrics[0]._key() + self.metrics[1]._key()

    def calculate(
        self,
        lmetric: AbstractMetric,
        rmetric: AbstractMetric,
    ) -> None:
        # value is scalar or ndarray
        values = [m.value for m in (lmetric, rmetric)]
        if self.is_proportional:
            self.value = min(values, key=np.linalg.norm)
        else:
            self.value = max(values, key=np.linalg.norm)
